#include "actmap.hh"

void ActMap::InitMap()
{
    map.clear();
    map.resize(As() * Cs() * Ms());
}

unsigned ActMap::MdInd(unsigned a, unsigned c, unsigned m) const
{
    return m + Ms() * (c + Cs() * a);
}

void ActMap::Set(unsigned a, unsigned c, unsigned m)
{
    map[MdInd(a, c, m)] = true;
}

#undef ELEMLIST_DEBUG
#include "elemlist"

template<typename collitemtype>
class ModeThing
{
    collitemtype base;
    const Lista<Mode> &list;
public:
    ModeThing(collitemtype b, const Lista<Mode> &l) : base(b), list(l) { }
    bool operator() (collitemtype c) const { return list[c-base].needsparm(); }
};

#include <list>
#include <vector>
#include <pair.h>
using std::list;
using std::pair;
using std::vector;

template<typename collitemtype>
class JobList
{
    typedef typename std::set<collitemtype>::const_iterator it;
    
    collitemtype agbase, chbase, mdbase;
    const Lista<Agent> &aglist;
    const Lista<Channel> &chlist;
    const Lista<Mode> &mdlist;
public:
    typedef list<pair<string, vector<collitemtype> > > joblisttype;
    joblisttype joblist;

    JobList(collitemtype a, collitemtype c, collitemtype m,
              const Lista<Agent> &ag, const Lista<Channel> &ch,
              const Lista<Mode> &md) : agbase(a), chbase(c), mdbase(m),
                aglist(ag), chlist(ch), mdlist(md) { }
    void operator()
    (
      const it ag1, const it ag2,
      const it ch1, const it ch2,
      const it md1, const it md2
    )
    {
        string s;
        for(it i = ch1; i != ch2; ++i) { if(s.size())s += ','; s += chlist[*i-chbase].getname(); }
        if(s.size())s += ' ';
        char sign=0;
        for(it i = md1; i != md2; ++i)
        {
            const Mode &m = mdlist[*i-mdbase];
            if(m.isnega()) { if(sign!='-')s += (sign='-'); }
            else { if(sign!='+')s += (sign='+'); }
            s += m.getmode();
        }
        for(it i = md1; i != md2; ++i)
        {
            const Mode &m = mdlist[*i-mdbase];
            if(m.needsparm())
            {
                s += ' ';
                s += m.getparm();
            }
        }
        
        vector<collitemtype> agents(ag1, ag2);
        for(unsigned a=0; a<agents.size(); ++a)agents[a] -= agbase;
        
        joblist.push_back(make_pair(s, agents));
    }
    
    void dump() const
    {
        typename joblisttype::const_iterator i;
        for(i=joblist.begin(); i!=joblist.end(); ++i)
        {
            cout << "Agents:";
            typename vector<collitemtype>::const_iterator j;
            for(j=i->second.begin(); j!=i->second.end(); ++j)
            {
                const Agent &a = aglist[*j];
                cout << ' ';
                a.dump();
            }
            cout << "\nJob: \"" << i->first << "\"\n\n";
        }
    }
    
    void distribute()
    {
        typedef typename joblisttype::const_iterator tit;
        
        unsigned combcount=0;        
        for(tit i=joblist.begin(); i!=joblist.end(); ++i)
            combcount += i->second.size();
        
        vector<unsigned> loads(aglist.Size());
        for(unsigned b=0; b<loads.size(); ++b)
            loads[b] = aglist[b].getload();
        
        unsigned bestload = 0, leastload = 0, bestcomb=0;
        for(unsigned a=0; a<combcount; ++a)
        {
            vector<unsigned> tmploads = loads;
            unsigned c=a;
            unsigned loadsum=0;
            for(tit i=joblist.begin(); i!=joblist.end(); ++i)
            {
                unsigned mod = i->second.size();
                unsigned agnum = c%mod; c/=mod;
                loadsum += ++tmploads[i->second[agnum]];
            }

            unsigned maxload = 0;
            for(unsigned b=0; b<tmploads.size(); ++b)
                if(tmploads[b] > maxload) maxload = tmploads[b];
            
            //cout << "Choice " << a << ": Load would be " << maxload << " (" << loadsum << ")\n";

            if(maxload < bestload || !a)
            {
                bestload = maxload;
            Good:
                leastload= loadsum;
                bestcomb = a; 
            }
            else if(maxload == bestload && loadsum < leastload)
            	goto Good;
        }
        
        for(tit i=joblist.begin(); i!=joblist.end(); ++i)
        {
            unsigned mod = i->second.size();
            unsigned agnum = bestcomb%mod; bestcomb/=mod;
            
            cout << "M " << aglist[i->second[agnum]].getname() << ' ' << i->first << endl;
        }
    }
};

void ActMap::Distribute()
{
#ifdef ELEMLIST_DEBUG
    typedef char collitemtype;
    collitemtype aBase = 'A';
    collitemtype cBase = 'A' + As();
    collitemtype mBase = 'A' + As() + Cs();
    
    cout << "Agents: ";
    for(unsigned a=0; a<As(); ++a) { cout << (collitemtype)(a+aBase) << '('; Agents[a].dump(); cout << ") "; }
    cout << "\nChannels: ";
    for(unsigned c=0; c<Cs(); ++c) { cout << (collitemtype)(c+cBase) << '('; Channels[c].dump(); cout << ") "; }
    cout << "\nModes: ";
    for(unsigned m=0; m<Ms(); ++m) { cout << (collitemtype)(m+mBase) << '('; Modes[m].dump(); cout << ") "; }
    cout << endl;
#else
    typedef unsigned collitemtype;
    collitemtype aBase=0, cBase=0, mBase=0;
#endif

    typedef elemlist<3, collitemtype> colltype;
    colltype coll;
    
    for(unsigned a=0; a<As(); ++a)
        for(unsigned c=0; c<Cs(); ++c)
            for(unsigned m=0; m<Ms(); ++m)
                if(map[MdInd(a,c,m)])
                {
                    colltype::Elem tmp;
                    tmp.insert(0, (collitemtype)(a+aBase));
                    tmp.insert(1, (collitemtype)(c+cBase));
                    tmp.insert(2, (collitemtype)(m+mBase));
                    coll.insert(tmp);
                }
    coll.optimize();
#ifdef ELEMLIST_DEBUG
    cout << "Optimized: ";
    coll.dump();
    cout << endl;
#endif
    coll.expand3(ModeThing<collitemtype> (mBase, Modes));
#ifdef ELEMLIST_DEBUG
    cout << "Expanded: ";
    coll.dump();
    cout << endl;
#endif
    coll.crop();
#ifdef ELEMLIST_DEBUG
    cout << "Cropped: ";
    coll.dump();
    cout << endl;
#endif
    
    JobList<collitemtype> jobs(
      aBase,cBase,mBase,
      Agents,Channels,Modes);
    coll.format(jobs);
    
#ifdef ELEMLIST_DEBUG
    coll.dump();
    cout << endl;
#endif

    jobs.distribute();
}

void ActMap::SetAgentChan(const Agent &ag, const Channel &ch)
{
    unsigned ai = AgentIndex(ag);
    unsigned ci = ChannelIndex(ch);
    
    agchan.Add(make_pair(ai, ci));
}

void ActMap::SetChanMode(const Channel &ch, const Mode &md)
{
    unsigned mi = ModeIndex(md);
    unsigned ci = ChannelIndex(ch);
    bool got = false;
    for(unsigned a=0; a<agchan.Size(); ++a)
    {
        pair<unsigned, unsigned> tmp = agchan[a];
        if(tmp.second == ci)
        {
            Set(tmp.first, ci, mi);
            got = true;
        }
    }
    if(!got)
        cout << "W " << ch.getname() << endl;
}
